@@ -4,8 +4,9 @@
 #   Copyright (C) 2013-2014 Sean Davis <smd.seandavis@gmail.com>
 #
 #   This program is free software: you can redistribute it and/or modify it
-#   under the terms of the GNU General Public License version 3, as published
-#   by the Free Software Foundation.
+#   under the terms of the GNU General Public License as published by
+#   the Free Software Foundation, either version 3 of the License, or
+#   (at your option) any later version.
 #
 #   This program is distributed in the hope that it will be useful, but
 #   WITHOUT ANY WARRANTY; without even the implied warranties of
@@ -191,9 +192,7 @@ class MugshotWindow(Window):
         self.fax_entry = builder.get_object('fax')
 
         # Users without sudo rights cannot change their name.
-        if not SudoDialog.check_sudo():
-            self.first_name_entry.set_sensitive(False)
-            self.last_name_entry.set_sensitive(False)
+        self.set_name_editable(SudoDialog.check_dependencies(['chfn']))
 
         # Stock photo browser
         self.stock_browser = builder.get_object('stock_browser')
@@ -216,6 +215,12 @@ class MugshotWindow(Window):
         # Populate all of the widgets.
         self.init_user_details()
 
+    def set_name_editable(self, editable):
+        """Set name fields editable."""
+        self.first_name_entry.set_sensitive(editable)
+        self.last_name_entry.set_sensitive(editable)
+        self.initials_entry.set_sensitive(editable)
+
     def init_user_details(self):
         """Initialize the user details entries and variables."""
         # Check for .face and set profile image.
@@ -223,55 +228,40 @@ class MugshotWindow(Window):
         face = os.path.join(home, '.face')
         logger.debug('Checking AccountsService for profile image')
         image = self.accounts_service_get_user_image()
-        logger.debug('Found profile image: %s' % str(image))
 
-        if os.path.isfile(face):
-            if os.path.samefile(image, face):
-                self.updated_image = face
+        # AccountsService may not be supported or desired.
+        if image is None:
+            logger.debug("AccountsService is not supported.")
+            self.updated_image = face
+            self.set_user_image(face)
+
+        # If it is supported, process and compare to ~/.face
+        else:
+            logger.debug('Found profile image: %s' % str(image))
+
+            if os.path.isfile(face):
+                if os.path.samefile(image, face):
+                    self.updated_image = face
+                else:
+                    self.updated_image = None
+                self.set_user_image(face)
+            elif os.path.isfile(image):
+                self.updated_image = image
+                self.set_user_image(image)
             else:
                 self.updated_image = None
-            self.set_user_image(face)
-        elif os.path.isfile(image):
-            self.updated_image = image
-            self.set_user_image(image)
-        else:
-            self.updated_image = None
-            self.set_user_image(None)
+                self.set_user_image(None)
 
-        # Search /etc/passwd for the current user's details.
-        logger.debug('Getting user details from /etc/passwd')
-        for line in open('/etc/passwd', 'r'):
-            if line.startswith(username + ':'):
-                logger.debug('Found details: %s' % line.strip())
-                details = line.split(':')[4]
-                name, office, office_phone, home_phone = details.split(',', 3)
-                break
-
-        # Expand the user's fullname into first and last.
-        try:
-            first_name, last_name = name.split(' ', 1)
-        except:
-            first_name = name
-            last_name = ''
-
-        # If the variables are defined as 'none', use blank for cleanliness.
-        if home_phone == 'none':
-            home_phone = ''
-        if office_phone == 'none':
-            office_phone = ''
-
-        # Get dconf settings
-        logger.debug('Getting initials, email, and fax from dconf')
-        initials = self.settings['initials']
-        email = self.settings['email']
-        fax = self.settings['fax']
+        user_details = self.get_user_details()
 
         # Set the class variables
-        self.first_name = first_name
-        self.last_name = last_name
-        self.initials = initials
-        self.home_phone = home_phone
-        self.office_phone = office_phone
+        self.first_name = user_details['first_name']
+        self.last_name = user_details['last_name']
+        self.initials = user_details['initials']
+        self.home_phone = user_details['home_phone']
+        self.office_phone = user_details['office_phone']
+        email = user_details['email']
+        fax = user_details['fax']
 
         # Populate the GtkEntries.
         logger.debug('Populating entries')
@@ -321,14 +311,24 @@ class MugshotWindow(Window):
         changes."""
         logger.debug('Applying changes...')
         if self.get_chfn_details_updated():
-            result = self.save_chfn_details()
-            if result != [0, 0]:
+            success, response = self.save_chfn_details()
+            if not success:
                 # Password was incorrect, complain.
-                primary = _("Authentication Failed")
+                if response in [Gtk.ResponseType.NONE,
+                                Gtk.ResponseType.CANCEL,
+                                Gtk.ResponseType.DELETE_EVENT]:
+                    msg_type = Gtk.MessageType.WARNING
+                    primary = _("Authentication cancelled.")
+                elif response == Gtk.ResponseType.REJECT:
+                    msg_type = Gtk.MessageType.WARNING
+                    primary = _("Authentication failed.")
+                else:
+                    msg_type = Gtk.MessageType.ERROR
+                    primary = _("An error occurred when saving changes.")
+
                 secondary = _("User details were not updated.")
                 dialog = Gtk.MessageDialog(transient_for=self, flags=0,
-                                           message_type=
-                                           Gtk.MessageType.WARNING,
+                                           message_type=msg_type,
                                            buttons=Gtk.ButtonsType.OK,
                                            text=primary)
                 dialog.format_secondary_text(secondary)
@@ -430,31 +430,34 @@ class MugshotWindow(Window):
 
     def accounts_service_get_user_image(self):
         """Get user profile image using AccountsService."""
-        bus = Gio.bus_get_sync(Gio.BusType.SYSTEM, None)
-        result = bus.call_sync('org.freedesktop.Accounts',
-                               '/org/freedesktop/Accounts',
-                               'org.freedesktop.Accounts',
-                               'FindUserByName',
-                               GLib.Variant('(s)', (username,)),
-                               GLib.VariantType.new('(o)'),
-                               Gio.DBusCallFlags.NONE,
-                               -1,
-                               None)
-        (path,) = result.unpack()
+        try:
+            bus = Gio.bus_get_sync(Gio.BusType.SYSTEM, None)
+            result = bus.call_sync('org.freedesktop.Accounts',
+                                   '/org/freedesktop/Accounts',
+                                   'org.freedesktop.Accounts',
+                                   'FindUserByName',
+                                   GLib.Variant('(s)', (username,)),
+                                   GLib.VariantType.new('(o)'),
+                                   Gio.DBusCallFlags.NONE,
+                                   -1,
+                                   None)
+            (path,) = result.unpack()
 
-        result = bus.call_sync('org.freedesktop.Accounts',
-                               path,
-                               'org.freedesktop.DBus.Properties',
-                               'GetAll',
-                               GLib.Variant('(s)',
-                                            ('org.freedesktop.Accounts.User',)
-                                            ),
-                               GLib.VariantType.new('(a{sv})'),
-                               Gio.DBusCallFlags.NONE,
-                               -1,
-                               None)
-        (props,) = result.unpack()
-        return props['IconFile']
+            variant = GLib.Variant('(s)',
+                                   ('org.freedesktop.Accounts.User',))
+            result = bus.call_sync('org.freedesktop.Accounts',
+                                   path,
+                                   'org.freedesktop.DBus.Properties',
+                                   'GetAll',
+                                   variant,
+                                   GLib.VariantType.new('(a{sv})'),
+                                   Gio.DBusCallFlags.NONE,
+                                   -1,
+                                   None)
+            (props,) = result.unpack()
+            return props['IconFile']
+        except GLib.GError:
+            return None
 
     def accounts_service_set_user_image(self, filename):
         """Set user profile image using AccountsService."""
@@ -558,31 +561,43 @@ class MugshotWindow(Window):
             logger.debug('chfn details have NOT been modified.')
             return False
 
+    def process_terminal_password(self, command, password):
+        """Handle password prompts from the interactive chfn commands."""
+        # Force the C language for guaranteed english strings in the script.
+        logger.debug('Executing: %s' % command)
+        child = SudoDialog.env_spawn(command, 5)
+        child.write_to_stdout = True
+        try:
+            child.expect([".*ssword.*", pexpect.EOF])
+            child.sendline(password)
+            child.expect([pexpect.EOF])
+        except pexpect.TIMEOUT:
+            logger.warning('Timeout reached, password was likely incorrect.')
+        child.close(True)
+        return child.exitstatus == 0
+
     def save_chfn_details(self):
         """Commit changes to chfn-related details.  For full name, changes must
         be performed as root.  Other changes are done with the user password.
 
-        Return exit codes for 1) full name changes and 2) home/work phone
-        changes.
-
-        e.g. [0, 0] (both passed)"""
-        return_codes = []
+        Return TRUE if successful."""
+        success = True
 
         # Get the password for sudo
         sudo_dialog = SudoDialog.SudoDialog(
             icon=None, name=_("Mugshot"), retries=3)
         sudo_dialog.format_primary_text(_("Enter your password to change user "
-                                        "details."))
+                                          "details."))
         sudo_dialog.format_secondary_text(_("This is a security measure to "
-                                          "prevent unwanted updates\n"
-                                          "to your personal information."))
-        sudo_dialog.run()
+                                            "prevent unwanted updates\n"
+                                            "to your personal information."))
+        response = sudo_dialog.run()
         sudo_dialog.hide()
         password = sudo_dialog.get_password()
         sudo_dialog.destroy()
 
         if not password:
-            return return_codes
+            return (False, response)
 
         sudo = which('sudo')
         chfn = which('chfn')
@@ -600,53 +615,30 @@ class MugshotWindow(Window):
             home_phone = 'none'
 
         # Full name can only be modified by root.  Try using sudo to modify.
-        if SudoDialog.check_sudo():
-            logger.debug('Attempting to set fullname with sudo chfn')
-            # Force the C language for guaranteed english strings in the script.
-            child = pexpect.spawn('%s %s %s' % (sudo, chfn, username),
-                                  env={"LANG": "C"})
-            child.timeout = 5
-            try:
-                child.expect([".*ssword.*", pexpect.EOF])
-                child.sendline(password)
-                child.expect(".*ame.*:")
-                child.sendline(full_name)
-                for i in range(5):
-                    child.sendline('')
-            except pexpect.TIMEOUT:
-                # Password was incorrect, or sudo rights not granted
-                logger.warning('Timeout reached, password was incorrect or '
-                               'sudo rights not granted.')
-                pass
-            child.close()
-            if child.exitstatus == 0:
+        if SudoDialog.check_dependencies(['chfn']):
+            logger.debug('Updating Full Name...')
+            command = "%s %s -f \"%s\" %s" % (sudo, chfn, full_name, username)
+            if self.process_terminal_password(command, password):
                 self.first_name = first_name
                 self.last_name = last_name
-            return_codes.append(child.exitstatus)
-        else:
-            return_codes.append(0)
+            else:
+                success = False
 
-        logger.debug('Attempting to set user details with chfn')
-        child = pexpect.spawn(chfn, env={"LANG": "C"})
-        child.timeout = 5
-        try:
-            child.expect([".*ssword.*", pexpect.EOF])
-            child.sendline(password)
-            child.expect(['Room Number.*:', 'Office.*:'])
-            child.sendline('')
-            child.expect(['Work Phone.*:', 'Office Phone.*:'])
-            child.sendline(office_phone)
-            child.expect('Home Phone.*:')
-            child.sendline(home_phone)
-            child.sendline(home_phone)
-        except pexpect.TIMEOUT:
-            logger.warning('Timeout reached, password was likely incorrect.')
-        child.close(True)
-        if child.exitstatus == 0:
-            self.office_phone = office_phone
+        logger.debug('Updating Home Phone...')
+        command = "%s -h \"%s\" %s" % (chfn, home_phone, username)
+        if self.process_terminal_password(command, password):
             self.home_phone = home_phone
-        return_codes.append(child.exitstatus)
-        return return_codes
+        else:
+            success = False
+
+        logger.debug('Updating Office Phone...')
+        command = "%s -w \"%s\" %s" % (chfn, office_phone, username)
+        if self.process_terminal_password(command, password):
+            self.office_phone = office_phone
+        else:
+            success = False
+
+        return (success, response)
 
     # = LibreOffice ========================================================= #
     def get_libreoffice_details_updated(self):
@@ -675,6 +667,83 @@ class MugshotWindow(Window):
         logger.debug('LibreOffice details do not need to be updated.')
         return False
 
+    def get_user_details(self):
+        """Use the various methods to get the most up-to-date version of the
+        user details."""
+        # Start with LibreOffice, as users may have configured that first.
+        data = self.get_libreoffice_data()
+
+        # Next is passwd, where we override name values with system values.
+        pwd_data = self.get_passwd_data()
+        data['first_name'] = pwd_data['first_name']
+        data['last_name'] = pwd_data['last_name']
+        data['initials'] = pwd_data['initials']
+        if len(data['home_phone']) == 0:
+            data['home_phone'] = pwd_data['home_phone']
+        if len(data['office_phone']) == 0:
+            data['office_phone'] = pwd_data['office_phone']
+
+        # Then get data from dconf
+        initials = self.settings['initials']
+        if len(initials) > 0:
+            data['initials'] = initials
+        email = self.settings['email']
+        if len(data['email']) == 0:
+            data['email'] = email
+        if len(data['fax']) == 0:
+            data['fax'] = self.settings['fax']
+
+        # Return all of the finalized information.
+        return data
+
+    def get_passwd_data(self):
+        """Get user details from passwd"""
+        # Use getent for current user's details.
+        try:
+            line = subprocess.check_output(['getent', 'passwd', username])
+            if isinstance(line, bytes):
+                line = line.decode('utf-8')
+            line = line.strip()
+            logger.debug('Found details: %s' % line.strip())
+            details = line.split(':')[4].split(',', 3)
+
+            while len(details) < 4:
+                details.append("")
+
+            # Extract the user details
+            name, office, office_phone, home_phone = details
+        except subprocess.CalledProcessError:
+            logger.warning("User %s not found in /etc/passwd. "
+                           "Mugshot may not function correctly." % username)
+            office = ""
+            office_phone = ""
+            home_phone = ""
+
+        # Use GLib to get the actual name.
+        name = GLib.get_real_name()
+
+        # Expand the user's fullname into first, last, and initials.
+        initials = name[0]
+        try:
+            first_name, last_name = name.split(' ', 1)
+            initials += last_name[0]
+        except:
+            first_name = name
+            last_name = ''
+
+        # If the variables are defined as 'none', use blank for cleanliness.
+        if home_phone == 'none':
+            home_phone = ''
+        if office_phone == 'none':
+            office_phone = ''
+
+        # Pack the data
+        data = {'first_name': first_name, 'last_name': last_name,
+                'home_phone': home_phone, 'office_phone': office_phone,
+                'initials': initials, 'email': '', 'fax': ''}
+
+        return data
+
     def get_libreoffice_data(self):
         """Get each of the preferences from the LibreOffice
         registymodifications preferences file.
